@@ -1,5 +1,5 @@
 """
-    Copyright (C) 2021  Michael Ablassmeier <abi@grinser.de>
+    Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -17,12 +17,11 @@
 import os
 import string
 import random
-import glob
 import logging
 from dataclasses import dataclass
 from socket import gethostname
 from argparse import Namespace
-from typing import Any, Dict, List, Tuple, Optional, Union
+from typing import Any, Dict, List, Tuple, Union
 from lxml.etree import _Element
 from lxml import etree as ElementTree
 import libvirt
@@ -31,8 +30,9 @@ from libvirtnbdbackup.virt.exceptions import (
     connectionFailed,
     startBackupFailed,
 )
-from libvirtnbdbackup import output
-from libvirtnbdbackup.output.exceptions import OutputException
+from libvirtnbdbackup.virt import fs
+from libvirtnbdbackup.virt import xml
+from libvirtnbdbackup.virt import disktype
 
 
 @dataclass
@@ -117,7 +117,7 @@ class client:
         return uri.startswith("qemu+ssh")
 
     def _useAuth(self, args: Namespace) -> bool:
-        """Check wether we want to use advanced auth method"""
+        """Check if we want to use advanced auth method"""
         if args.uri.startswith("qemu+"):
             return True
         if self._reqAuth(args.uri):
@@ -130,6 +130,9 @@ class client:
     def _connect(self, args: Namespace) -> libvirt.virConnect:
         """return libvirt connection handle"""
         log.debug("Libvirt URI: [%s]", args.uri)
+        localHostname = gethostname()
+        log.debug("Hostname: [%s]", localHostname)
+
         if self._useAuth(args):
             log.debug(
                 "Login information specified, connect libvirtd using openAuth function."
@@ -146,13 +149,16 @@ class client:
                 conn = self._connectAuth(args.uri, args.user, args.password)
             else:
                 conn = self._connectOpen(args.uri)
-            if gethostname() != conn.getHostname():
+
+            remoteHostname = conn.getHostname()
+            log.debug("Hostname returned by libvirt API: [%s]", remoteHostname)
+            if localHostname != remoteHostname:
                 log.info(
                     "Connected to remote host: [%s], local host: [%s]",
                     conn.getHostname(),
                     gethostname(),
                 )
-                self.remoteHost = conn.getHostname()
+                self.remoteHost = remoteHostname
 
             return conn
 
@@ -160,11 +166,6 @@ class client:
 
         return self._connectOpen(args.uri)
 
-    @staticmethod
-    def _getTree(vmConfig: str) -> _Element:
-        """Return Etree element for vm config"""
-        return ElementTree.fromstring(vmConfig)
-
     def getDomain(self, name: str) -> libvirt.virDomain:
         """Lookup domain"""
         try:
@@ -216,7 +217,7 @@ class client:
         if self.libvirtVersion >= 7006000:
             return True
 
-        tree = self._getTree(domObj.XMLDesc(0))
+        tree = xml.asTree(domObj.XMLDesc(0))
         for target in tree.findall(
             "{http://libvirt.org/schemas/domain/qemu/1.0}capabilities"
         ):
@@ -260,7 +261,7 @@ class client:
     def getDomainInfo(self, vmConfig: str) -> Dict[str, str]:
         """Return object with general vm information relevant
         for backup"""
-        tree = self._getTree(vmConfig)
+        tree = xml.asTree(vmConfig)
         settings = {}
 
         for flag in ["loader", "nvram", "kernel", "initrd"]:
@@ -275,7 +276,7 @@ class client:
     def adjustDomainConfigRemoveDisk(self, vmConfig: str, excluded) -> bytes:
         """Remove disk from config, in case it has been excluded
         from the backup."""
-        tree = self._getTree(vmConfig)
+        tree = xml.asTree(vmConfig)
         log.info("Removing excluded disk [%s] from vm config.", excluded)
         try:
             target = tree.xpath(f"devices/disk/target[@dev='{excluded}']")[0]
@@ -290,9 +291,9 @@ class client:
         self, args: Namespace, restoreDisk: DomainDisk, vmConfig: str, targetFile: str
     ) -> bytes:
         """Adjust virtual machine configuration after restoring. Changes
-        the pathes to the virtual machine disks and attempts to remove
+        the paths to the virtual machine disks and attempts to remove
         components excluded during restore."""
-        tree = self._getTree(vmConfig)
+        tree = xml.asTree(vmConfig)
 
         try:
             log.info("Removing uuid setting from vm config.")
@@ -313,7 +314,30 @@ class client:
             if disk.get("type") == "volume":
                 log.info("Disk has type volume, resetting to type file.")
                 disk.set("type", "file")
+
             dev = disk.xpath("target")[0].get("dev")
+            log.debug("Handling target device: [%s]", dev)
+
+            device = disk.get("device")
+            driver = disk.xpath("driver")[0].get("type")
+
+            if disktype.Optical(device, dev):
+                log.info("Removing device [%s], type [%s] from vm config", dev, device)
+                disk.getparent().remove(disk)
+                continue
+
+            if disktype.Raw(driver, device):
+                log.warning(
+                    "Removing raw disk [%s] from vm config, use --raw to copy as is.",
+                    dev,
+                )
+                disk.getparent().remove(disk)
+                continue
+            backingStore = disk.xpath("backingStore")
+            if backingStore:
+                log.info("Removing existent backing store settings")
+                disk.remove(backingStore[0])
+
             originalFile = disk.xpath("source")[0].get("file")
             if dev == restoreDisk.target:
                 log.info(
@@ -323,23 +347,6 @@ class client:
                     targetFile,
                 )
                 disk.xpath("source")[0].set("file", targetFile)
-            device = disk.get("device")
-            driver = disk.xpath("driver")[0].get("type")
-            if device in ("lun", "cdrom", "floppy"):
-                log.info("Removing [%s] device from vm config", device)
-                disk.getparent().remove(disk)
-                continue
-            if driver == "raw" and args.raw is False:
-                log.warning(
-                    "Removing raw disk [%s] from vm config.",
-                    dev,
-                )
-                disk.getparent().remove(disk)
-                continue
-            backingStore = disk.xpath("backingStore")
-            if backingStore:
-                log.info("Removing existant backing store settings")
-                disk.remove(backingStore[0])
 
         return ElementTree.tostring(tree, encoding="utf8", method="xml")
 
@@ -382,7 +389,7 @@ class client:
         """Parse virtual machine configuration for disk devices, filter
         all non supported devices
         """
-        tree = self._getTree(vmConfig)
+        tree = xml.asTree(vmConfig)
         devices = []
 
         excludeList = None
@@ -391,52 +398,41 @@ class client:
 
         for disk in tree.xpath("devices/disk"):
             dev = disk.xpath("target")[0].get("dev")
+            device = disk.get("device")
+            diskFormat = disk.xpath("driver")[0].get("type")
 
             if excludeList is not None and dev in excludeList:
                 log.warning("Excluding disk [%s] from backup as requested", dev)
                 continue
 
-            # ignore attached lun or direct access block devices
-            if disk.xpath("target")[0].get("type") == "block":
-                log.warning(
-                    "Device [%s] does not support changed block tracking, skipping.",
-                    dev,
-                )
+            # skip cdrom/floppy devices
+            if disktype.Optical(device, dev):
                 continue
 
-            device = disk.get("device")
-            if device == "lun":
-                log.warning(
-                    "Skipping direct attached lun [%s].",
-                    dev,
-                )
-                continue
-            if device in ("cdrom", "floppy"):
-                log.info("Skipping attached [%s] device: [%s].", device, dev)
-                continue
-
-            # ignore disk which use raw format, they do not support CBT
-            diskFormat = disk.xpath("driver")[0].get("type")
-            if diskFormat == "raw" and args.raw is False:
-                log.warning(
-                    "Raw disk [%s] excluded by default, use option --raw to include.",
-                    dev,
-                )
+            # include other direct attached devices if --raw option is enabled
+            if args.raw is False and (
+                disktype.Block(disk, dev)
+                or disktype.Lun(device, dev)
+                or disktype.Raw(diskFormat, dev)
+            ):
                 continue
 
+            diskPath = None
             diskType = disk.get("type")
             if diskType == "volume":
                 log.debug("Disk [%s]: volume notation", dev)
                 diskPath = self._getDiskPathByVolume(disk)
             elif diskType == "file":
-                log.debug("Disk [%s]: file notation")
+                log.debug("Disk [%s]: file notation", dev)
                 diskPath = disk.xpath("source")[0].get("file")
             elif diskType == "block":
-                log.warning(
-                    "Skipping direct attached block device [%s].",
-                    dev,
-                )
-                continue
+                if args.raw is False:
+                    log.warning(
+                        "Skipping direct attached block device [%s], use option --raw to include.",
+                        dev,
+                    )
+                    continue
+                diskPath = disk.xpath("source")[0].get("dev")
             else:
                 log.error("Unable to detect disk volume type for disk [%s]", dev)
                 continue
@@ -464,24 +460,6 @@ class client:
         log.debug("Device list: %s ", devices)
         return devices
 
-    @staticmethod
-    def _indentXml(top: _Element) -> str:
-        """Indent xml output for debug log"""
-        try:
-            ElementTree.indent(top)
-        except ElementTree.ParseError as errmsg:
-            log.debug("Failed to parse xml: [%s]", errmsg)
-        except AttributeError:
-            # older ElementTree verisons dont have the
-            # indent method, skip silently and use
-            # non formatted string
-            pass
-
-        xml = ElementTree.tostring(top).decode()
-        log.debug("\n%s", xml)
-
-        return xml
-
     def _createBackupXml(self, args: Namespace, diskList) -> str:
         """Create XML file for starting an backup task using libvirt API."""
         top = ElementTree.Element("domainbackup", {"mode": "pull"})
@@ -517,9 +495,7 @@ class client:
             dE = ElementTree.SubElement(disks, "disk", {"name": disk.target})
             ElementTree.SubElement(dE, "scratch", {"file": f"{scratchFile}"})
 
-        xml = self._indentXml(top)
-
-        return xml
+        return xml.indent(top)
 
     def _createCheckpointXml(
         self, diskList: List[Any], parentCheckpoint: str, checkpointName: str
@@ -540,40 +516,16 @@ class client:
             # because it is not supported. Backup will only be crash
             # consistent. If we would like to create a consistent
             # backup, we would have to create an snapshot for these
-            # kind of disks.
+            # kind of disks, example:
+            # virsh checkpoint-create-as vm4 --diskspec sdb
+            # error: unsupported configuration:  \
+            # checkpoint for disk sdb unsupported for storage type raw
+            # See also:
+            # https://lists.gnu.org/archive/html/qemu-devel/2021-03/msg07448.html
             if disk.format != "raw":
                 ElementTree.SubElement(disks, "disk", {"name": disk.target})
 
-        xml = self._indentXml(top)
-
-        return xml
-
-    @staticmethod
-    def fsFreeze(domObj: libvirt.virDomain, mountpoints: None) -> bool:
-        """Attempt to freeze domain filesystems using qemu guest agent"""
-        log.debug("Attempting to freeze filesystems.")
-        try:
-            if mountpoints is not None:
-                frozen = domObj.fsFreeze(mountpoints.split(","))
-            else:
-                frozen = domObj.fsFreeze()
-            log.info("Freezed [%s] filesystems.", frozen)
-            return True
-        except libvirt.libvirtError as errmsg:
-            log.warning(errmsg)
-            return False
-
-    @staticmethod
-    def fsThaw(domObj: libvirt.virDomain) -> bool:
-        """Thaw freeze filesystems"""
-        log.debug("Attempting to thaw filesystems.")
-        try:
-            thawed = domObj.fsThaw()
-            log.info("Thawed [%s] filesystems.", thawed)
-            return True
-        except libvirt.libvirtError as errmsg:
-            log.warning(errmsg)
-            return False
+        return xml.indent(top)
 
     def startBackup(
         self,
@@ -592,7 +544,7 @@ class client:
             checkpointXml = self._createCheckpointXml(
                 diskList, args.cpt.parent, args.cpt.name
             )
-        freezed = self.fsFreeze(domObj, args.freeze_mountpoint)
+        freezed = fs.freeze(domObj, args.freeze_mountpoint)
         try:
             log.debug("Starting backup job via libvirt API.")
             domObj.backupBegin(backupXml, checkpointXml)
@@ -608,91 +560,7 @@ class client:
             # check if filesystem is freezed and thaw
             # in case creating checkpoint fails.
             if freezed is True:
-                self.fsThaw(domObj)
-
-    @staticmethod
-    def _checkpointExists(
-        domObj: libvirt.virDomain, checkpointName: str
-    ) -> libvirt.virDomainCheckpoint:
-        """Check if an checkpoint exists"""
-        return domObj.checkpointLookupByName(checkpointName)
-
-    @staticmethod
-    def _getCheckpointXml(cptObj: libvirt.virDomainCheckpoint) -> str:
-        """Get Checkpoint XML including size, if possible. Flag
-        is not supported amongst all libvirt versions."""
-        try:
-            return cptObj.getXMLDesc(libvirt.VIR_DOMAIN_CHECKPOINT_XML_SIZE)
-        except libvirt.libvirtError as e:
-            log.warning("Failed to get checkpoint info with size information: [%s]", e)
-            return cptObj.getXMLDesc()
-
-    def getCheckpointSize(self, domObj: libvirt.virDomain, checkpointName: str) -> int:
-        """Return current size of checkpoint for all disks"""
-        size = 0
-        cpt = self._checkpointExists(domObj, checkpointName)
-        cptTree = self._getTree(self._getCheckpointXml(cpt))
-        for s in cptTree.xpath("disks/disk/@size"):
-            size += int(s)
-
-        return size
-
-    def removeAllCheckpoints(
-        self,
-        domObj: libvirt.virDomain,
-        checkpointList: Union[List[Any], None],
-        args: Namespace,
-        defaultCheckpointName: str,
-    ) -> bool:
-        """Remove all existing checkpoints for a virtual machine,
-        used during FULL backup to reset checkpoint chain
-        """
-        log.debug("Cleaning up persistent storage %s", args.checkpointdir)
-        try:
-            for checkpointFile in glob.glob(f"{args.checkpointdir}/*.xml"):
-                log.debug("Remove checkpoint file: %s", checkpointFile)
-                os.remove(checkpointFile)
-        except OSError as e:
-            log.error(
-                "Failed to clean persistent storage %s: %s", args.checkpointdir, e
-            )
-            return False
-
-        if checkpointList is None:
-            cpts = domObj.listAllCheckpoints()
-            if cpts:
-                for cpt in cpts:
-                    if self._deleteCheckpoint(cpt, defaultCheckpointName) is False:
-                        return False
-            return True
-
-        for checkpoint in checkpointList:
-            cptObj = self._checkpointExists(domObj, checkpoint)
-            if cptObj:
-                if self._deleteCheckpoint(cptObj, defaultCheckpointName) is False:
-                    return False
-        return True
-
-    @staticmethod
-    def _deleteCheckpoint(
-        cptObj: libvirt.virDomainCheckpoint, defaultCheckpointName: str
-    ) -> bool:
-        """Delete checkpoint"""
-        checkpointName = cptObj.getName()
-        if defaultCheckpointName not in checkpointName:
-            log.debug(
-                "Skipping checkpoint removal: [%s]: not from this application",
-                checkpointName,
-            )
-            return True
-        log.debug("Attempt to remove checkpoint: [%s]", checkpointName)
-        try:
-            cptObj.delete()
-            log.debug("Removed checkpoint: [%s]", checkpointName)
-            return True
-        except libvirt.libvirtError as errmsg:
-            log.error("Error during checkpoint removal: [%s]", errmsg)
-            return False
+                fs.thaw(domObj)
 
     @staticmethod
     def stopBackup(domObj: libvirt.virDomain) -> bool:
@@ -703,92 +571,3 @@ class client:
         except libvirt.libvirtError as err:
             log.warning("Failed to stop backup job: [%s]", err)
             return False
-
-    def redefineCheckpoints(self, domObj: libvirt.virDomain, args: Namespace) -> bool:
-        """Redefine checkpoints from persistent storage"""
-        log.info("Loading checkpoint list from: [%s]", args.checkpointdir)
-        checkpointList = glob.glob(f"{args.checkpointdir}/*.xml")
-        checkpointList.sort(key=os.path.getmtime)
-
-        for checkpointFile in checkpointList:
-            log.debug("Loading checkpoint config from: [%s]", checkpointFile)
-            try:
-                with output.openfile(checkpointFile, "rb") as f:
-                    checkpointConfig = f.read()
-                    root = ElementTree.fromstring(checkpointConfig)
-            except OutputException as e:
-                log.error("Opening checkpoint file failed: [%s]: %s", checkpointFile, e)
-                return False
-            except ElementTree.ParseError as e:
-                log.error(
-                    "Failed to load checkpoint config from [%s]: %s", checkpointFile, e
-                )
-                return False
-
-            try:
-                checkpointName = root.find("name").text
-            except ElementTree.ParseError as e:
-                log.error("Failed to find checkpoint name: [%s]", e)
-                return False
-
-            try:
-                _ = self._checkpointExists(domObj, checkpointName)
-                log.debug("Checkpoint [%s] found", checkpointName)
-                continue
-            except libvirt.libvirtError as e:
-                # ignore VIR_ERR_NO_DOMAIN_CHECKPOINT, report other errors
-                if e.get_error_code() != libvirt.VIR_ERR_NO_DOMAIN_CHECKPOINT:
-                    log.error("libvirt error: %s", e)
-                    return False
-
-            log.info("Redefine missing checkpoint: [%s]", checkpointName)
-            try:
-                domObj.checkpointCreateXML(
-                    checkpointConfig.decode(),
-                    libvirt.VIR_DOMAIN_CHECKPOINT_CREATE_REDEFINE,
-                )
-            except libvirt.libvirtError as e:
-                log.error("Redefining checkpoint failed: [%s]: %s", checkpointName, e)
-                return False
-
-        return True
-
-    def backupCheckpoint(self, args: Namespace, domObj: libvirt.virDomain) -> bool:
-        """save checkpoint config to persistent storage"""
-        checkpointFile = f"{args.checkpointdir}/{args.cpt.name}.xml"
-        log.info("Saving checkpoint config to: [%s]", checkpointFile)
-        try:
-            with output.openfile(checkpointFile, "wb") as f:
-                c = self._checkpointExists(domObj, args.cpt.name)
-                f.write(self._getCheckpointXml(c).encode())
-                return True
-        except OutputException as errmsg:
-            log.error(
-                "Failed to save checkpoint config to file: [%s]: %s",
-                checkpointFile,
-                errmsg,
-            )
-            return False
-
-    @staticmethod
-    def hasforeignCheckpoint(
-        domObj: libvirt.virDomain, defaultCheckpointName: str
-    ) -> Optional[str]:
-        """Check if the virtual machine has an checkpoint which was not
-        created by virtnbdbackup
-
-        If an user or a third party utility creates an checkpoint,
-        it is in line with the complete checkpoint chain, but
-        virtnbdbackup does not save it. We can ensure consistency
-        only if the complete chain of checkpoints is created by
-        ourself. In case we detect an checkpoint that does not
-        match our name, return so.
-        """
-        cpts = domObj.listAllCheckpoints()
-        if cpts:
-            for cpt in cpts:
-                checkpointName = cpt.getName()
-                log.debug("Found foreign checkpoint: [%s]", checkpointName)
-                if defaultCheckpointName not in checkpointName:
-                    return checkpointName
-        return None
